import pytest

from whatlies.language import FasttextLanguage


@pytest.fixture()
def lang():
    return FasttextLanguage("tests/custom_fasttext_model.bin")


@pytest.mark.parametrize(
    "text", [("red red", "blue red"), ("red", "green", "blue"), ("dog", "cat")]
)
def test_check_sizes(lang, text):
    X = text
    assert lang.fit(X).transform(X).shape == (len(text), 10)
    assert lang.fit_transform(X).shape == (len(text), 10)


def test_get_params(lang):
    assert "model" in lang.get_params().keys()
    assert "size" in lang.get_params().keys()



from whatlies.language import SpacyLanguage

import pytest


@pytest.fixture()
def embset():
    lang = SpacyLanguage("en_core_web_md")
    names = [
        "red",
        "blue",
        "green",
        "yellow",
        "cat",
        "dog",
        "mouse",
        "rat",
        "bike",
        "car",
    ]
    return lang[names]


def test_plot_distance_raises_error(embset):
    with pytest.raises(ValueError):
        embset.plot_distance(metric="dinosaurhead")


def test_plot_similarity_raises_error(embset):
    with pytest.raises(ValueError):
        embset.plot_distance(metric="dinosaurhead")



from whatlies import Embedding, EmbeddingSet
from whatlies.helpers import reverse_strings


def test_reverse_strings():
    embset = EmbeddingSet(Embedding(name="helloworld", vector=[1, 2])).pipe(
        reverse_strings
    )
    emb = [e for e in embset][0]
    assert emb.name == "dlrowolleh"



import pytest

import numpy as np

from whatlies.language import SpacyLanguage
from whatlies.transformers import Pca


words = [
    "prince",
    "princess",
    "nurse",
    "doctor",
    "banker",
    "man",
    "woman",
    "cousin",
    "neice",
    "king",
    "queen",
    "dude",
    "guy",
    "gal",
    "fire",
    "dog",
    "cat",
    "mouse",
    "red",
    "blue",
    "green",
    "yellow",
    "water",
    "person",
    "family",
    "brother",
    "sister",
]

# I'm loading in the spaCy model globally because it is much faster this way.
lang = SpacyLanguage("en_core_web_md")


@pytest.fixture
def embset():
    return lang[words]


def test_set_title_works(embset):
    ax = embset.plot_3d(annot=True, title="foobar")
    assert ax.title._text == "foobar"


def test_correct_points_plotted(embset):
    embset_plt = embset.transform(Pca(3))
    ax = embset_plt.plot_3d(annot=True)
    offset = ax.collections[0]._offsets3d
    assert np.all(np.array(offset).T == embset_plt.to_X())


def test_correct_points_plotted_mapped(embset):
    embset_plt = embset.transform(Pca(3))
    ax = embset_plt.plot_3d("king", "red", "dog", annot=True)
    offset = ax.collections[0]._offsets3d
    king, red, dog = [v for v in np.array(offset)]

    assert np.all(king == np.array([embset_plt[w] > embset_plt["king"] for w in words]))
    assert np.all(red == np.array([embset_plt[w] > embset_plt["red"] for w in words]))
    assert np.all(dog == np.array([embset_plt[w] > embset_plt["dog"] for w in words]))


def test_basic_dimensions_3d_chart(embset):
    embset_plt = embset.transform(Pca(3))
    ax = embset_plt.plot_3d(annot=True, title="foobar")
    assert ax.xaxis.get_label_text() == "Dimension 0"
    assert ax.yaxis.get_label_text() == "Dimension 1"
    assert ax.zaxis.get_label_text() == "Dimension 2"
    assert [t.get_text() for t in ax.texts] == words


def test_named_dimensions_3d_chart(embset):
    ax = embset.transform(Pca(3)).plot_3d("king", "queen", "prince", annot=True)
    assert ax.xaxis.get_label_text() == "king"
    assert ax.yaxis.get_label_text() == "queen"
    assert ax.zaxis.get_label_text() == "prince"
    assert [t.get_text() for t in ax.texts] == words


def test_named_dimensions_3d_chart_rename(embset):
    ax = embset.transform(Pca(3)).plot_3d(
        "king", "queen", "prince", annot=True, x_label="x", y_label="y"
    )
    assert ax.xaxis.get_label_text() == "x"
    assert ax.yaxis.get_label_text() == "y"
    assert ax.zaxis.get_label_text() == "prince"
    assert [t.get_text() for t in ax.texts] == words



from umap import UMAP

from whatlies.transformers._transformer import SklearnTransformer


class Umap(SklearnTransformer):
    """
    This transformer transformers all vectors in an [EmbeddingSet][whatlies.embeddingset.EmbeddingSet]
    by means of umap. We're using the implementation in [umap-learn](https://umap-learn.readthedocs.io/en/latest/).

    Arguments:
        n_components: the number of compoments to create/add
        kwargs: keyword arguments passed to the UMAP algorithm

    Usage:

    ```python
    from whatlies.language import SpacyLanguage
    from whatlies.transformers import Umap

    words = ["prince", "princess", "nurse", "doctor", "banker", "man", "woman",
             "cousin", "neice", "king", "queen", "dude", "guy", "gal", "fire",
             "dog", "cat", "mouse", "red", "blue", "green", "yellow", "water",
             "person", "family", "brother", "sister"]

    lang = SpacyLanguage("en_core_web_md")
    emb = lang[words]

    emb.transform(Umap(3)).plot_interactive_matrix(0, 1, 2)
    ```
    """

    def __init__(self, n_components=2, **kwargs):
        super().__init__(
            UMAP, f"umap_{n_components}", n_components=n_components, **kwargs
        )



class NotInstalled:
    """
    This object is used for optional dependencies. If a backend is not installed we
    replace the transformer/language with this object. This allows us to give a friendly
    message to the user that they need to install extra dependencies as well as a link
    to our documentation page.
    """

    def __init__(self, tool, dep):
        self.tool = tool
        self.dep = dep

        msg = f"In order to use {self.tool} you'll need to install via;\n\n"
        msg += f"pip install whatlies[{self.dep}]\n\n"
        msg += "See installation guide here: https://rasahq.github.io/whatlies/#installation."
        self.msg = msg

    def __getattr__(self, *args, **kwargs):
        raise ModuleNotFoundError(self.msg)

    def __call__(self, *args, **kwargs):
        raise ModuleNotFoundError(self.msg)



import pytest

import numpy as np
from sklearn.preprocessing import normalize

from whatlies.language import SpacyLanguage
from whatlies.transformers import Noise, AddRandom, Normalizer, Umap, Tsne, Pca
from whatlies.transformers._transformer import Transformer, SklearnTransformer

words = [
    "prince",
    "princess",
    "nurse",
    "doctor",
    "banker",
    "man",
    "woman",
    "cousin",
    "neice",
    "king",
    "queen",
    "dude",
    "guy",
    "gal",
    "fire",
    "dog",
    "cat",
    "mouse",
    "red",
    "blue",
    "green",
    "yellow",
    "water",
    "person",
    "family",
    "brother",
    "sister",
]

lang = SpacyLanguage("en_core_web_sm")
emb = lang[words]

transformers = [
    Umap(2),
    Umap(3),
    Pca(2),
    Pca(3),
    Noise(0.1),
    Noise(0.01),
    AddRandom(n=4),
    AddRandom(n=1),
    lambda d: d | (d["man"] - d["woman"]),
    Tsne(2, n_iter=250),
    Tsne(3, n_iter=250),
    Normalizer(),
    Normalizer(feature=True),
]
extra_sizes = [0, 0, 0, 0, 0, 0, 4, 1, 0, 0, 0, 0, 0, 0, 0]
tfm_ids = [_.__class__.__name__ for _ in transformers]


@pytest.mark.parametrize(
    "transformer,extra_size", zip(transformers, extra_sizes), ids=tfm_ids
)
def test_transformations_new_size(transformer, extra_size):
    emb_new = emb.transform(transformer)
    assert len(emb_new) == len(emb) + extra_size


@pytest.mark.parametrize(
    "transformer",
    [
        Umap(2),
        Pca(2),
        Noise(0.1),
        Tsne(2, n_iter=250),
        AddRandom(n=4),
        lambda d: d | (d["man"] - d["woman"]),
        Normalizer(),
    ],
)
def test_transformations_keep_props(transformer):
    emb_new = emb.add_property("group", lambda d: "one").transform(transformer)
    for w in words:
        assert hasattr(emb_new[w], "group")


@pytest.mark.parametrize(
    "transformer",
    [
        Umap(2),
        Pca(2),
        Noise(0.1),
        AddRandom(n=4),
        Tsne(2, n_iter=250),
        Normalizer(),
    ],
)
def test_transformers_are_subclassed_properly(transformer):
    assert isinstance(transformer, Transformer)


@pytest.mark.parametrize(
    "transformer",
    [
        Umap(2),
        Pca(2),
        Tsne(2, n_iter=250),
    ],
)
def test_sklearn_transformers_are_subclassed_properly(transformer):
    assert isinstance(transformer, SklearnTransformer)


def test_transformer_base_class():
    with pytest.raises(TypeError, match="Can't instantiate abstract class Transformer"):
        Transformer()


def test_normalizer_transformer():
    X = emb.to_X()
    normalized_emb = emb.transform(Normalizer())
    assert set(emb.embeddings.keys()) == set(normalized_emb.embeddings.keys())
    assert np.array_equal(normalized_emb.to_X(), normalize(X, norm="l1"))

    normalized_emb = emb.transform(Normalizer(norm="l2"))
    assert set(emb.embeddings.keys()) == set(normalized_emb.embeddings.keys())
    assert np.array_equal(normalized_emb.to_X(), normalize(X))

    normalized_emb = emb.transform(Normalizer(feature=True))
    assert set(emb.embeddings.keys()) == set(normalized_emb.embeddings.keys())
    assert np.array_equal(normalized_emb.to_X(), normalize(X, norm="l1", axis=0))



import pytest

from whatlies.language import TFHubLanguage


@pytest.fixture
def lang(request):
    return TFHubLanguage(**request.param)


@pytest.mark.parametrize(
    "lang, expected_shape",
    [
        (
            {"url": "https://tfhub.dev/google/tf2-preview/gnews-swivel-20dim/1"},
            (20,),
        ),
        (
            {
                "url": "https://tfhub.dev/google/tf2-preview/gnews-swivel-20dim-with-oov/1"
            },
            (20,),
        ),
    ],
    indirect=["lang"],
)
def test_basic_usage(lang, expected_shape):
    emb = lang["a test sentence"]
    assert emb.vector.shape == expected_shape
    emb = lang[["test", "a simple test sentence", "and another nice sentence"]]
    assert len(emb) == 3
    assert emb["test"].vector.shape == expected_shape



import pytest
import numpy as np
import matplotlib as mpl
import scipy.spatial.distance as scipy_distance

from whatlies import Embedding, EmbeddingSet
from .common import validate_plot_general_properties


"""
*Guide*

Here are the plot's propertites which could be checked (some of them may not be applicable
for a particular plot):
    - type: the class type of collection in matplotlib to ensure the right kind of plot
        has been created.
    - data: the position of points, arrows or texts in the plot, depending on the plot's type.
    - x_label: label of x-axis.
    - y_label: label of y-axis.
    - tilte: title of the plot.
    - aspect: aspect ratio of plot, usually 'auto' unless `axis_option` argument is set.
    - color: color of points (in scatter plot) or arrows (in arrow plot). It should be rgba values.
    - label: label of points (in scatter plot) or arrows (in arrow plot).
"""


@pytest.fixture
def embset():
    names = ["red", "blue", "green", "yellow", "white"]
    vectors = np.random.rand(5, 3)
    embeddings = [Embedding(name, vector) for name, vector in zip(names, vectors)]
    return EmbeddingSet(*embeddings)


def test_embeddingset_plot_scatter_str_axis(embset):
    fig, ax = mpl.pyplot.subplots()
    embset.plot(kind="scatter", x_axis="blue", y_axis="red")
    vectors = []
    for emb in embset.embeddings.values():
        vec = []
        vec.append(emb > embset["blue"])
        vec.append(emb > embset["red"])
        vectors.append(vec)
    vectors = np.array(vectors)
    props = {
        "type": mpl.collections.PathCollection,
        "data": vectors,
        "x_label": "blue",
        "y_label": "red",
        "title": "",
        "label": list(embset.embeddings.keys()),
        "color": mpl.colors.to_rgba_array("steelblue"),
        "aspect": "auto",
    }
    assert isinstance(ax.collections[0], props["type"])
    assert np.array_equal(ax.collections[0].get_offsets(), props["data"])
    assert [t.get_text() for t in ax.texts] == props["label"]
    assert np.array_equal(ax.collections[0].get_facecolors(), props["color"])
    validate_plot_general_properties(ax, props)
    mpl.pyplot.close(fig)


def test_embeddingset_plot_arrow_str_axis(embset):
    fig, ax = mpl.pyplot.subplots()
    embset.plot(
        kind="arrow",
        x_axis="blue",
        y_axis="red",
        x_label="blue vec",
        y_label="red vec",
        title="test plot",
        color="yellow",
    )
    vectors = []
    for emb in embset.embeddings.values():
        vec = []
        vec.append(emb > embset["blue"])
        vec.append(emb > embset["red"])
        vectors.append(vec)
    vectors = np.array(vectors)
    props = {
        "type": mpl.collections.PolyCollection,
        "data": vectors,
        "x_label": "blue vec",
        "y_label": "red vec",
        "title": "test plot",
        "label": list(embset.embeddings.keys()),
        "color": mpl.colors.to_rgba_array("yellow"),
        "aspect": "auto",
    }
    UV = np.concatenate(
        (ax.collections[1].U[:, None], ax.collections[1].V[:, None]), axis=-1
    )
    assert isinstance(ax.collections[1], props["type"])
    assert np.array_equal(UV, props["data"])
    assert [t.get_text() for t in ax.texts] == props["label"]
    assert np.array_equal(ax.collections[1].get_facecolors(), props["color"])
    validate_plot_general_properties(ax, props)
    mpl.pyplot.close(fig)


def test_embeddingset_plot_scatter_integer_axis(embset):
    fig, ax = mpl.pyplot.subplots()
    embset.plot(kind="scatter", x_axis=0, y_axis=1, annot=False, color="black")
    vectors = np.concatenate((embset.to_X()[:, 0:1], embset.to_X()[:, 1:2]), axis=-1)
    props = {
        "type": mpl.collections.PathCollection,
        "data": vectors,
        "x_label": "Dimension 0",
        "y_label": "Dimension 1",
        "title": "",
        "color": mpl.colors.to_rgba_array("black"),
        "aspect": "auto",
        # Not applicable: label
    }
    assert isinstance(ax.collections[0], props["type"])
    assert np.array_equal(ax.collections[0].get_offsets(), props["data"])
    assert ax.texts == []
    assert np.array_equal(ax.collections[0].get_facecolors(), props["color"])
    validate_plot_general_properties(ax, props)
    mpl.pyplot.close(fig)


def test_embeddingset_plot_arrow_integer_axis(embset):
    fig, ax = mpl.pyplot.subplots()
    embset.plot(
        kind="arrow",
        x_axis=1,
        y_axis=2,
        x_label="1",
        y_label="2",
        color="yellow",
        axis_option="scaled",
    )
    vectors = np.concatenate((embset.to_X()[:, 1:2], embset.to_X()[:, 2:3]), axis=-1)
    props = {
        "type": mpl.collections.PolyCollection,
        "data": vectors,
        "x_label": "1",
        "y_label": "2",
        "title": "",
        "label": list(embset.embeddings.keys()),
        "color": mpl.colors.to_rgba_array("yellow"),
        "aspect": 1.0,
    }
    UV = np.concatenate(
        (ax.collections[1].U[:, None], ax.collections[1].V[:, None]), axis=-1
    )
    assert isinstance(ax.collections[1], props["type"])
    assert np.array_equal(UV, props["data"])
    assert [t.get_text() for t in ax.texts] == props["label"]
    assert np.array_equal(ax.collections[1].get_facecolors(), props["color"])
    validate_plot_general_properties(ax, props)
    mpl.pyplot.close(fig)


def test_embeddingset_plot_scatter_emb_axis(embset):
    fig, ax = mpl.pyplot.subplots()
    embset.plot(kind="scatter", x_axis=embset["green"], y_axis=embset["white"])
    vectors = []
    for emb in embset.embeddings.values():
        vec = []
        vec.append(emb > embset["green"])
        vec.append(emb > embset["white"])
        vectors.append(vec)
    vectors = np.array(vectors)
    props = {
        "type": mpl.collections.PathCollection,
        "data": vectors,
        "x_label": "green",
        "y_label": "white",
        "title": "",
        "label": list(embset.embeddings.keys()),
        "color": mpl.colors.to_rgba_array("steelblue"),
        "aspect": "auto",
    }
    assert isinstance(ax.collections[0], props["type"])
    assert np.array_equal(ax.collections[0].get_offsets(), props["data"])
    assert [t.get_text() for t in ax.texts] == props["label"]
    assert np.array_equal(ax.collections[0].get_facecolors(), props["color"])
    validate_plot_general_properties(ax, props)
    mpl.pyplot.close(fig)


def test_embeddingset_plot_arrow_emb_axis(embset):
    fig, ax = mpl.pyplot.subplots()
    embset.plot(
        kind="arrow",
        x_axis=embset["blue"],
        y_axis=embset["red"],
        x_label="xx",
        color="magenta",
    )
    vectors = []
    for emb in embset.embeddings.values():
        vec = []
        vec.append(emb > embset["blue"])
        vec.append(emb > embset["red"])
        vectors.append(vec)
    vectors = np.array(vectors)
    props = {
        "type": mpl.collections.PolyCollection,
        "data": vectors,
        "x_label": "xx",
        "y_label": "red",
        "title": "",
        "label": list(embset.embeddings.keys()),
        "color": mpl.colors.to_rgba_array("magenta"),
        "aspect": "auto",
    }
    UV = np.concatenate(
        (ax.collections[1].U[:, None], ax.collections[1].V[:, None]), axis=-1
    )
    assert isinstance(ax.collections[1], props["type"])
    assert np.array_equal(UV, props["data"])
    assert [t.get_text() for t in ax.texts] == props["label"]
    assert np.array_equal(ax.collections[1].get_facecolors(), props["color"])
    validate_plot_general_properties(ax, props)
    mpl.pyplot.close(fig)


def test_embeddingset_plot_arrow_integer_axis_with_str_axis_metric(embset):
    fig, ax = mpl.pyplot.subplots()
    embset.plot(
        kind="arrow",
        x_axis=1,
        y_axis=2,
        axis_metric="cosine_distance",
        x_label="1",
        y_label="2",
        color="yellow",
        axis_option="scaled",
    )
    vectors = np.concatenate((embset.to_X()[:, 1:2], embset.to_X()[:, 2:3]), axis=-1)
    props = {
        "type": mpl.collections.PolyCollection,
        "data": vectors,
        "x_label": "1",
        "y_label": "2",
        "title": "",
        "label": list(embset.embeddings.keys()),
        "color": mpl.colors.to_rgba_array("yellow"),
        "aspect": 1.0,
    }
    UV = np.concatenate(
        (ax.collections[1].U[:, None], ax.collections[1].V[:, None]), axis=-1
    )
    assert isinstance(ax.collections[1], props["type"])
    assert np.array_equal(UV, props["data"])
    assert [t.get_text() for t in ax.texts] == props["label"]
    assert np.array_equal(ax.collections[1].get_facecolors(), props["color"])
    validate_plot_general_properties(ax, props)
    mpl.pyplot.close(fig)


def test_embeddingset_plot_scatter_emb_axis_with_common_str_axis_metric(embset):
    fig, ax = mpl.pyplot.subplots()
    embset.plot(
        kind="scatter",
        x_axis=embset["green"],
        y_axis=embset["white"],
        axis_metric="euclidean",
    )
    vectors = []
    for emb in embset.embeddings.values():
        vectors.append(
            [
                scipy_distance.euclidean(emb.vector, embset["green"].vector),
                scipy_distance.euclidean(emb.vector, embset["white"].vector),
            ]
        )
    vectors = np.array(vectors)
    props = {
        "type": mpl.collections.PathCollection,
        "data": vectors,
        "x_label": "green",
        "y_label": "white",
        "title": "",
        "label": list(embset.embeddings.keys()),
        "color": mpl.colors.to_rgba_array("steelblue"),
        "aspect": "auto",
    }
    assert isinstance(ax.collections[0], props["type"])
    assert np.array_equal(ax.collections[0].get_offsets(), props["data"])
    assert [t.get_text() for t in ax.texts] == props["label"]
    assert np.array_equal(ax.collections[0].get_facecolors(), props["color"])
    validate_plot_general_properties(ax, props)
    mpl.pyplot.close(fig)


def test_embeddingset_plot_arrow_emb_axis_with_different_axis_metric(embset):
    fig, ax = mpl.pyplot.subplots()
    embset.plot(
        kind="arrow",
        x_axis=embset["blue"],
        y_axis="red",
        axis_metric=[scipy_distance.correlation, "cosine_similarity"],
        x_label="xx",
        color="magenta",
    )
    vectors = []
    for emb in embset.embeddings.values():
        vectors.append(
            [
                scipy_distance.correlation(emb.vector, embset["blue"].vector),
                1.0 - scipy_distance.cosine(emb.vector, embset["red"].vector),
            ]
        )
    vectors = np.array(vectors)
    props = {
        "type": mpl.collections.PolyCollection,
        "data": vectors,
        "x_label": "xx",
        "y_label": "red",
        "title": "",
        "label": list(embset.embeddings.keys()),
        "color": mpl.colors.to_rgba_array("magenta"),
        "aspect": "auto",
    }
    UV = np.concatenate(
        (ax.collections[1].U[:, None], ax.collections[1].V[:, None]), axis=-1
    )
    assert isinstance(ax.collections[1], props["type"])
    assert np.array_equal(UV, props["data"])
    assert [t.get_text() for t in ax.texts] == props["label"]
    assert np.array_equal(ax.collections[1].get_facecolors(), props["color"])
    validate_plot_general_properties(ax, props)
    mpl.pyplot.close(fig)


def test_embeddingset_plot_scatter_mixed_axis(embset):
    fig, ax = mpl.pyplot.subplots()
    embset.plot(kind="scatter", x_axis=embset["green"], y_axis=2)
    vectors = []
    for emb in embset.embeddings.values():
        vec = []
        vec.append(emb > embset["green"])
        vec.append(emb.vector[2])
        vectors.append(vec)
    vectors = np.array(vectors)
    props = {
        "type": mpl.collections.PathCollection,
        "data": vectors,
        "x_label": "green",
        "y_label": "Dimension 2",
        "title": "",
        "label": list(embset.embeddings.keys()),
        "color": mpl.colors.to_rgba_array("steelblue"),
        "aspect": "auto",
    }
    assert isinstance(ax.collections[0], props["type"])
    assert np.array_equal(ax.collections[0].get_offsets(), props["data"])
    assert [t.get_text() for t in ax.texts] == props["label"]
    assert np.array_equal(ax.collections[0].get_facecolors(), props["color"])
    validate_plot_general_properties(ax, props)
    mpl.pyplot.close(fig)


def test_embeddingset_plot_arrow_mixed_axis(embset):
    fig, ax = mpl.pyplot.subplots()
    embset.plot(kind="arrow", x_axis=1, y_axis="red", x_label="xx", color="magenta")
    vectors = []
    for emb in embset.embeddings.values():
        vec = []
        vec.append(emb.vector[1])
        vec.append(emb > embset["red"])
        vectors.append(vec)
    vectors = np.array(vectors)
    props = {
        "type": mpl.collections.PolyCollection,
        "data": vectors,
        "x_label": "xx",
        "y_label": "red",
        "title": "",
        "label": list(embset.embeddings.keys()),
        "color": mpl.colors.to_rgba_array("magenta"),
        "aspect": "auto",
    }
    UV = np.concatenate(
        (ax.collections[1].U[:, None], ax.collections[1].V[:, None]), axis=-1
    )
    assert isinstance(ax.collections[1], props["type"])
    assert np.array_equal(UV, props["data"])
    assert [t.get_text() for t in ax.texts] == props["label"]
    assert np.array_equal(ax.collections[1].get_facecolors(), props["color"])
    validate_plot_general_properties(ax, props)
    mpl.pyplot.close(fig)


def test_embeddingset_plot_raises_error_when_str_axis_not_exists(embset):
    with pytest.raises(KeyError):
        embset.plot(x_axis="bnb", y_axis="blue")
    with pytest.raises(KeyError):
        embset.plot(x_axis="red", y_axis="clk")



import numpy as np
from sklearn.preprocessing import FunctionTransformer

from ._transformer import SklearnTransformer


class Noise(SklearnTransformer):
    """
    This transformer adds gaussian noise to an embeddingset.

    Arguments:
        sigma: the amount of gaussian noise to add
        seed: seed value for random number generator

    Usage:

    ```python
    from whatlies.language import SpacyLanguage
    from whatlies.transformers import Noise

    words = ["prince", "princess", "nurse", "doctor", "banker", "man", "woman",
             "cousin", "neice", "king", "queen", "dude", "guy", "gal", "fire",
             "dog", "cat", "mouse", "red", "blue", "green", "yellow", "water",
             "person", "family", "brother", "sister"]

    lang = SpacyLanguage("en_core_web_md")
    emb = lang[words]

    emb.transform(Noise(3))
    ```
    """

    def __init__(self, sigma=0.1, seed=42, **kwargs):
        np.random.seed(seed)
        super().__init__(
            FunctionTransformer,
            f"noise_{sigma}",
            func=lambda X: X + np.random.normal(0, sigma, X.shape),
            **kwargs,
        )



import pytest
import numpy as np
import matplotlib as mpl
import scipy.spatial.distance as scipy_distance

from whatlies import Embedding, EmbeddingSet
from .common import validate_plot_general_properties


"""
*Guide*

Here are the plot's propertites which could be checked (some of them may not be applicable
for a particular plot):
    - type: the class type of collection in matplotlib to ensure the right kind of plot
        has been created.
    - data: the position of points, arrows or texts in the plot, depending on the plot's type.
    - x_label: label of x-axis.
    - y_label: label of y-axis.
    - tilte: title of the plot.
    - aspect: aspect ratio of plot, usually 'auto' unless `axis_option` argument is set.
    - color: color of points (in scatter plot) or arrows (in arrow plot). It should be rgba values.
    - label: label of points (in scatter plot) or arrows (in arrow plot).
"""


@pytest.fixture
def embset():
    names = ["red", "blue", "green", "yellow", "white"]
    vectors = np.random.rand(5, 3)
    embeddings = [Embedding(name, vector) for name, vector in zip(names, vectors)]
    return EmbeddingSet(*embeddings)


def test_embedding_plot_scatter_integer_axis(embset):
    emb = embset["red"]
    fig, ax = mpl.pyplot.subplots()
    emb.plot(kind="scatter", x_axis=0, y_axis=1)
    props = {
        "type": mpl.collections.PathCollection,
        "data": emb.vector[0:2],
        "x_label": "Dimension 0",
        "y_label": "Dimension 1",
        "title": "",
        "color": mpl.colors.to_rgba_array("steelblue"),
        "label": "red",
        "aspect": "auto",
    }
    assert np.array_equal(ax.collections[0].get_offsets()[0], props["data"])
    assert isinstance(ax.collections[0], props["type"])
    assert np.array_equal(ax.collections[0].get_facecolor(), props["color"])
    assert ax.texts[0].get_text() == props["label"]
    validate_plot_general_properties(ax, props)
    mpl.pyplot.close(fig)


def test_embedding_plot_arrow_integer_axis(embset):
    emb = embset["red"]
    fig, ax = mpl.pyplot.subplots()
    emb.plot(
        kind="arrow",
        x_axis=0,
        y_axis=2,
        color="blue",
        x_label="xlabel",
        y_label="ylabel",
        title="test plot",
        annot=False,
    )
    props = {
        "type": mpl.collections.PolyCollection,
        "data": np.concatenate((emb.vector[0:1], emb.vector[2:3])),
        "x_label": "xlabel",
        "y_label": "ylabel",
        "title": "test plot",
        "color": mpl.colors.to_rgba_array("blue"),
        "aspect": "auto",
        # Not applicable: label
    }
    UV = np.concatenate((ax.collections[1].U, ax.collections[1].V))
    assert isinstance(ax.collections[1], props["type"])
    assert np.array_equal(UV, props["data"])
    assert np.array_equal(ax.collections[1].get_facecolor(), props["color"])
    assert ax.texts == []
    validate_plot_general_properties(ax, props)
    mpl.pyplot.close(fig)


def test_embedding_plot_text_integer_axis(embset):
    emb = embset["red"]
    fig, ax = mpl.pyplot.subplots()
    emb.plot(kind="text", x_axis=1, y_axis=2)
    props = {
        "data": np.concatenate((emb.vector[1:2] + 0.01, emb.vector[2:3])),
        "x_label": "Dimension 1",
        "y_label": "Dimension 2",
        "title": "",
        "label": "red",
        "aspect": "auto",
        # Not applicable: type, color
    }
    assert np.array_equal(ax.texts[0].get_position(), props["data"])
    assert ax.collections == []
    assert ax.texts[0].get_text() == props["label"]
    validate_plot_general_properties(ax, props)
    mpl.pyplot.close(fig)


def test_embedding_plot_scatter_emb_axis(embset):
    emb = embset["red"]
    fig, ax = mpl.pyplot.subplots()
    emb.plot(kind="scatter", x_axis=embset["blue"], y_axis=embset["green"])
    props = {
        "type": mpl.collections.PathCollection,
        "data": np.array([emb > embset["blue"], emb > embset["green"]]),
        "x_label": "blue",
        "y_label": "green",
        "color": mpl.colors.to_rgba_array("steelblue"),
        "title": "",
        "label": "red",
        "aspect": "auto",
    }
    assert np.array_equal(ax.collections[0].get_offsets()[0], props["data"])
    assert isinstance(ax.collections[0], props["type"])
    assert ax.texts[0].get_text() == props["label"]
    validate_plot_general_properties(ax, props)
    mpl.pyplot.close(fig)


def test_embedding_plot_arrow_emb_axis(embset):
    emb = embset["red"] + embset["yellow"]
    fig, ax = mpl.pyplot.subplots()
    emb.plot(
        kind="arrow",
        x_axis=embset["blue"],
        y_axis=embset["green"],
        color="yellow",
        show_ops=True,
        axis_option="equal",
    )
    props = {
        "type": mpl.collections.PolyCollection,
        "data": np.array([emb > embset["blue"], emb > embset["green"]]),
        "x_label": "blue",
        "y_label": "green",
        "color": mpl.colors.to_rgba_array("yellow"),
        "title": "",
        "label": "(red + yellow)",
        "aspect": 1.0,
    }
    UV = np.concatenate((ax.collections[1].U, ax.collections[1].V))
    assert isinstance(ax.collections[1], props["type"])
    assert np.array_equal(UV, props["data"])
    assert np.array_equal(ax.collections[1].get_facecolor(), props["color"])
    assert ax.texts[0].get_text() == props["label"]
    validate_plot_general_properties(ax, props)
    mpl.pyplot.close(fig)


def test_embedding_plot_arrow_integer_axis_with_str_axis_metric(embset):
    emb = embset["red"]
    fig, ax = mpl.pyplot.subplots()
    emb.plot(
        kind="arrow",
        x_axis=0,
        y_axis=2,
        axis_metric="euclidean",
        color="blue",
        x_label="xlabel",
        y_label="ylabel",
        title="test plot",
        annot=False,
    )
    props = {
        "type": mpl.collections.PolyCollection,
        "data": np.concatenate((emb.vector[0:1], emb.vector[2:3])),
        "x_label": "xlabel",
        "y_label": "ylabel",
        "title": "test plot",
        "color": mpl.colors.to_rgba_array("blue"),
        "aspect": "auto",
        # Not applicable: label
    }
    UV = np.concatenate((ax.collections[1].U, ax.collections[1].V))
    assert isinstance(ax.collections[1], props["type"])
    assert np.array_equal(UV, props["data"])
    assert np.array_equal(ax.collections[1].get_facecolor(), props["color"])
    assert ax.texts == []
    validate_plot_general_properties(ax, props)
    mpl.pyplot.close(fig)


def test_embedding_plot_scatter_emb_axis_with_common_str_axis_metric(embset):
    emb = embset["red"]
    fig, ax = mpl.pyplot.subplots()
    emb.plot(
        kind="scatter",
        x_axis=embset["blue"],
        y_axis=embset["green"],
        axis_metric="cosine_distance",
    )
    props = {
        "type": mpl.collections.PathCollection,
        "data": np.array(
            [
                scipy_distance.cosine(emb.vector, embset["blue"].vector),
                scipy_distance.cosine(emb.vector, embset["green"].vector),
            ]
        ),
        "x_label": "blue",
        "y_label": "green",
        "color": mpl.colors.to_rgba_array("steelblue"),
        "title": "",
        "label": "red",
        "aspect": "auto",
    }
    assert np.array_equal(ax.collections[0].get_offsets()[0], props["data"])
    assert isinstance(ax.collections[0], props["type"])
    assert ax.texts[0].get_text() == props["label"]
    validate_plot_general_properties(ax, props)
    mpl.pyplot.close(fig)


def test_embedding_plot_arrow_emb_axis_with_different_str_axis_metric(embset):
    emb = embset["red"] + embset["yellow"]
    fig, ax = mpl.pyplot.subplots()
    emb.plot(
        kind="arrow",
        x_axis=embset["blue"],
        y_axis=embset["green"],
        axis_metric=["euclidean", "cosine_similarity"],
        color="yellow",
        show_ops=True,
        axis_option="equal",
    )
    props = {
        "type": mpl.collections.PolyCollection,
        "data": np.array(
            [
                scipy_distance.euclidean(emb.vector, embset["blue"].vector),
                1.0 - scipy_distance.cosine(emb.vector, embset["green"].vector),
            ]
        ),
        "x_label": "blue",
        "y_label": "green",
        "color": mpl.colors.to_rgba_array("yellow"),
        "title": "",
        "label": "(red + yellow)",
        "aspect": 1.0,
    }
    UV = np.concatenate((ax.collections[1].U, ax.collections[1].V))
    assert isinstance(ax.collections[1], props["type"])
    assert np.array_equal(UV, props["data"])
    assert np.array_equal(ax.collections[1].get_facecolor(), props["color"])
    assert ax.texts[0].get_text() == props["label"]
    validate_plot_general_properties(ax, props)
    mpl.pyplot.close(fig)


def test_embedding_plot_arrow_emb_axis_with_callable_y_axis_metric(embset):
    emb = embset["red"] + embset["yellow"]
    fig, ax = mpl.pyplot.subplots()
    emb.plot(
        kind="arrow",
        x_axis=embset["blue"],
        y_axis=embset["green"],
        axis_metric=[None, scipy_distance.correlation],
        x_label="xaxis",
        y_label="corr",
        color="yellow",
        show_ops=True,
        axis_option="equal",
    )
    props = {
        "type": mpl.collections.PolyCollection,
        "data": np.array(
            [
                emb > embset["blue"],
                scipy_distance.correlation(emb.vector, embset["green"].vector),
            ]
        ),
        "x_label": "xaxis",
        "y_label": "corr",
        "color": mpl.colors.to_rgba_array("yellow"),
        "title": "",
        "label": "(red + yellow)",
        "aspect": 1.0,
    }
    UV = np.concatenate((ax.collections[1].U, ax.collections[1].V))
    assert isinstance(ax.collections[1], props["type"])
    assert np.array_equal(UV, props["data"])
    assert np.array_equal(ax.collections[1].get_facecolor(), props["color"])
    assert ax.texts[0].get_text() == props["label"]
    validate_plot_general_properties(ax, props)
    mpl.pyplot.close(fig)


def test_embedding_plot_raises_error_when_incorrect_axis_type(embset):
    emb = embset["red"]
    with pytest.raises(ValueError, match="The `x_axis` value should be"):
        emb.plot(x_axis=1.0)
    with pytest.raises(ValueError, match="The `y_axis` value should be"):
        emb.plot(y_axis="blue")


def test_embedding_plot_raises_error_when_incorrect_axis_metric(embset):
    emb = embset["red"]
    with pytest.raises(ValueError, match="The given axis metric is not supported"):
        emb.plot(x_axis=embset["blue"], axis_metric="correlation")
    with pytest.raises(ValueError, match="The given axis metric type is not"):
        emb.plot(y_axis=embset["blue"], axis_metric=1)



import pathlib

import pytest
import fasttext

from whatlies.language import FasttextLanguage

text_path = str(pathlib.Path(__file__).parent.parent.absolute() / "data" / "foobar.txt")
model1 = fasttext.train_unsupervised(
    text_path, model="cbow", dim=20, epoch=20, min_count=1
)
model2 = fasttext.train_unsupervised(text_path, model="skipgram", dim=10, min_count=1)


def test_load_in_model1():
    lang = FasttextLanguage(model1)
    assert lang["dog"].vector.shape[0] == 20


def test_load_in_model2():
    lang = FasttextLanguage(model2)
    assert lang["dog"].vector.shape[0] == 10


def test_retreive_similar_len():
    assert len(FasttextLanguage(model1).score_similar("cat", 20)) == 20
    assert len(FasttextLanguage(model2).score_similar("cat", 10)) == 10
    assert len(FasttextLanguage(model1).score_similar("cat", 1000)) == 91
    assert len(FasttextLanguage(model2).score_similar("cat", 1000)) == 91


def test_raise_warning():
    with pytest.warns(UserWarning):
        FasttextLanguage(model1).score_similar("cat", 1000)



import pytest
import numpy as np

from whatlies import Embedding, EmbeddingSet


@pytest.fixture
def emb():
    x = Embedding("x", [0.0, 1.0])
    y = Embedding("y", [1.0, 0.0])
    z = Embedding("z", [0.5, 0.5])
    return EmbeddingSet(x, y, z)


def test_emb_dist(emb):
    assert np.isclose(emb["x"].distance(emb["x"]), 0.0)
    assert np.isclose(emb["x"].distance(emb["y"], metric="euclidean"), np.sqrt(2))
    assert np.isclose(emb["x"].distance(emb["z"], metric="euclidean"), np.sqrt(2) / 2)


def test_emb_norm(emb):
    assert np.isclose(emb["x"].norm, 1.0)
    assert np.isclose(emb["y"].norm, 1.0)
    assert np.isclose(emb["z"].norm, np.sqrt(0.5))


def test_emb_add(emb):
    new_emb = emb["x"] + emb["y"]
    assert np.isclose(new_emb.vector, np.array([1.0, 1.0])).all()


def test_emb_subtract(emb):
    new_emb = emb["x"] - emb["y"]
    assert np.isclose(new_emb.vector, np.array([-1.0, 1.0])).all()


def test_emb_proj_unto1(emb):
    new_emb = emb["z"] >> emb["y"]
    assert np.isclose(new_emb.vector, np.array([0.5, 0.0])).all()


def test_emb_proj_unto2(emb):
    new_emb = emb["z"] >> emb["x"]
    assert np.isclose(new_emb.vector, np.array([0.0, 0.5])).all()


def test_emb_proj_away1(emb):
    new_emb = emb["z"] | emb["y"]
    assert np.isclose(new_emb.vector, np.array([0.0, 0.5])).all()


def test_emb_proj_away2(emb):
    new_emb = emb["z"] | emb["x"]
    assert np.isclose(new_emb.vector, np.array([0.5, 0.0])).all()


def test_emb_gt(emb):
    assert pytest.approx(emb["z"] > emb["x"], 1.0)
    assert pytest.approx(emb["x"] > emb["z"], 0.5)


def test_emb_plot_no_err_2d(emb):
    emb["x"].plot(kind="arrow").plot(kind="text")
    emb["y"].plot(kind="arrow").plot(kind="text")
    emb["z"].plot(kind="arrow").plot(kind="text")


def test_emb_plot_no_err_3d():
    x = Embedding("x", [0.0, 1.0, 1.0])
    y = Embedding("y", [1.0, 0.0, 1.0])
    z = Embedding("z", [0.5, 0.5, 1.0])
    for item in [x, y, z]:
        item.plot("scatter", x_axis=x, y_axis=y)


def test_emb_str_method(emb):
    for char in "xyz":
        assert str(emb[char]) == char


def test_emb_ndim():
    foo = Embedding("foo", [0, 1, 0.2])
    assert foo.ndim == 3


def test_negation():
    foo = Embedding("foo", [0.1, 0.3])
    assert np.allclose((-foo).vector, -np.array([0.1, 0.3]))



from sklearn.preprocessing import normalize, FunctionTransformer

from ._transformer import SklearnTransformer


class Normalizer(SklearnTransformer):
    """
    This transformer normalizes the embeddings in an `EmbeddingSet` instance.

    Arguments:
        norm: the normalization value, could be either of `'l1'`, `'l2'` or `'max'`.
        feature: if `True` each feature (i.e. dimension) of embeddings would be normalized
            independently to unit norm; otherwise, each embedding vector would be normalized to unit norm.

    Usage:

    ```python
    from whatlies.language import SpacyLanguage
    from whatlies.transformers import Normalizer

    words = ["prince", "princess", "nurse", "doctor", "banker", "man", "woman",
             "cousin", "neice", "king", "queen", "dude", "guy", "gal", "fire",
             "dog", "cat", "mouse", "red", "blue", "green", "yellow", "water",
             "person", "family", "brother", "sister"]

    lang = SpacyLanguage("en_core_web_md")
    emb = lang[words]

    emb.transform(Normalizer(norm='l1'))
    emb.transform(Normalizer(feature=True))
    ```
    """

    def __init__(self, norm: str = "l1", feature: bool = False, **kwargs) -> None:
        self.norm = norm
        self.feature = feature
        super().__init__(
            FunctionTransformer,
            f"norm_{norm}",
            func=lambda X: normalize(X, norm=self.norm, axis=0 if self.feature else 1),
            **kwargs,
        )



import os
import sys

import numpy as np
from sklearn.utils.validation import check_is_fitted
from sklearn.base import BaseEstimator, TransformerMixin


class SklearnTransformerMixin(BaseEstimator, TransformerMixin):
    def fit(self, X, y=None):
        """
        Will fit the language model such that it is ready for use in scikit learn.
        Check out the [guide](https://rasahq.github.io/whatlies/tutorial/languages/#scikit-learn) for more details.
        """
        if not np.array(X).dtype.type is np.str_:
            raise ValueError("You must give this preprocessor text as input.")
        self.fitted_ = True
        return self

    def partial_fit(self, X, y=None):
        """
        No-op.
        """
        if not np.array(X).dtype.type is np.str_:
            raise ValueError("You must give this preprocessor text as input.")
        self.fitted_ = True
        return self

    def transform(self, X):
        """
        Will apply the transformer as if it is a scikit-learn text feature extractor.
        Check out the [guide](https://rasahq.github.io/whatlies/tutorial/languages/#scikit-learn) for more details.
        """
        check_is_fitted(self, "fitted_")
        if not np.array(X).dtype.type is np.str_:
            raise ValueError("You must give this preprocessor text as input.")
        return np.array([np.nan_to_num(self[x].vector) for x in X])


class HiddenPrints:
    def __enter__(self):
        self._original_stdout = sys.stderr
        sys.stderr = open(os.devnull, "w")

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stderr.close()
        sys.stderr = self._original_stdout



from whatlies import Embedding, EmbeddingSet


def reverse_strings(embset):
    """
    This helper will reverse the strings in the embeddingset. This can be useful
    for making matplotlib plots with Arabic texts.

    This helper is meant to be used via `EmbeddingSet.pipe()`.

    Arguments:
        embset: EmbeddingSet to adapt

    Usage:

    ```python
    from whatlies.helpers import reverse_strings
    from whatlies.language import BytePairLanguage

    translation = {
       "man":"رجل",
       "woman":"امرأة",
       "king":"ملك",
       "queen":"ملكة",
       "brother":"أخ",
       "sister":"أخت",
       "cat":"قطة",
       "dog":"كلب",
       "lion":"أسد",
       "puppy":"جرو",
       "male student":"طالب",
       "female student":"طالبة",
       "university":"جامعة",
       "school":"مدرسة",
       "kitten":" قطة صغيرة",
       "apple" : "تفاحة",
       "orange" : "برتقال",
       "cabbage" : "كرنب",
       "carrot" : "جزرة"
    }

    lang_cv  = BytePairLanguage("ar")

    arabic_words = list(words.values())

    # before
    lang_cv[translation].plot_similarity()

    # after
    lang_cv[translation].pipe(reverse_strings).plot_similarity()
    ```

    ![](https://rasahq.github.io/whatlies/images/arabic-before-after.png)
    """
    return EmbeddingSet(
        *[Embedding(name=e.name[::-1], vector=e.vector) for e in embset]
    )



def validate_plot_general_properties(ax, props):
    assert ax.xaxis.get_label_text() == props["x_label"]
    assert ax.yaxis.get_label_text() == props["y_label"]
    assert ax.get_title() == props["title"]
    assert ax.get_aspect() == props["aspect"]



from typing import Union, List, Optional

import tensorflow_text  # noqa: F401
import tensorflow_hub as tfhub

from whatlies.embedding import Embedding
from whatlies.embeddingset import EmbeddingSet
from whatlies.language._common import SklearnTransformerMixin


class TFHubLanguage(SklearnTransformerMixin):
    """
    This class provides the abitilty to load and use text-embedding models of Tensorflow Hub to
    retrieve [Embedding][whatlies.embedding.Embedding]s or [EmbeddingSet][whatlies.embeddingset.EmbeddingSet]s from them.
    A list of supported models is available [here](https://tfhub.dev/s?module-type=text-embedding&tf-version=tf2);
    however, note that only those models which operate directly on raw text (i.e. don't require any pre-processing
    such as tokenization) are supported for the moment (e.g. models such as BERT or ALBERT are not supported).
    Further, the TF-Hub compatible models from other repositories (i.e. other than [tfhub.dev](https://tfhub.dev/))
    are also supported.

    Important:
        This object will automatically download a large file if it is not cached yet.

        This language model does not contain a vocabulary, so it cannot be used
        to retreive similar tokens. Use an `EmbeddingSet` instead.

        This language backend might require you to manually install extra dependencies
        unless you installed via either;

        ```
        pip install whatlies[tfhub]
        pip install whatlies[all]
        ```

        Further, consider that this language model mainly supports TensorFlow 2.x models (i.e. TF2 `SavedModel` format);
        although, TensorFlow 1.x models might be supported to some extent as well (see
        [`hub.load`](https://www.tensorflow.org/hub/api_docs/python/hub/load) documentation as well as
        [model compatibility guide](https://www.tensorflow.org/hub/model_compatibility)).

    Arguments:
        url: The url or local directory path of the model.
        tags: A set of strings specifying the graph variant to use, if loading from a TF1 module.
            It is passed to `hub.load` function.
        signature: An optional signature of the model to use.

    **Usage**:

    ```python
    > from whatlies.language import TFHubLanguage
    > lang = TFHubLanguage("https://tfhub.dev/google/nnlm-en-dim50/2")
    > lang['today is a gift']
    > lang = TFHubLanguage("https://tfhub.dev/google/nnlm-en-dim50/2")
    > lang[['withdraw some money', 'take out cash', 'cash out funds']]
    ```
    """

    def __init__(
        self,
        url: str,
        tags: Optional[List[str]] = None,
        signature: Optional[str] = None,
    ) -> None:
        model = tfhub.load(url, tags=tags)
        if signature:
            model = model.signatures[signature]
        self.signature = signature
        self.url = url
        self.tags = tags
        self.model = model

    def __getitem__(
        self, query: Union[str, List[str]]
    ) -> Union[Embedding, EmbeddingSet]:
        """
        Retreive a single embedding or a set of embeddings.

        Arguments:
            query: single string or list of strings

        **Usage**

        ```python
        > from whatlies.language import TFHubLanguage
        > lang = TFHubLanguage("https://tfhub.dev/google/nnlm-en-dim50/2")
        > lang['today is a gift']
        > lang = TFHubLanguage("https://tfhub.dev/google/nnlm-en-dim50/2")
        > lang[['withdraw some money', 'take out cash', 'cash out funds']]
        ```
        """
        if isinstance(query, str):
            return self._get_embedding(query)
        return EmbeddingSet(*[self._get_embedding(q) for q in query])

    def _get_embedding(self, query: str) -> Embedding:
        vec = self.model([query]).numpy()[0]
        return Embedding(query, vec)



from pathlib import Path

import pytest
import numpy as np
from sklearn.pipeline import Pipeline
from sklearn.pipeline import FeatureUnion
from sklearn.linear_model import LogisticRegression
from sklearn.feature_extraction.text import CountVectorizer

from whatlies.language import DIETLanguage


@pytest.mark.rasa
def test_sklearn_pipeline_works():
    lang = DIETLanguage(
        model_path=next(Path("tests/rasa-test-demo/models").glob("*.tar.gz"))
    )
    pipe = Pipeline([("embed", lang), ("model", LogisticRegression())])

    X = [
        "i really like this post",
        "thanks for that comment",
        "i enjoy this friendly forum",
        "this is a bad post",
        "this is a bad post",
        "i dislike this article",
        "this is not well written",
    ]
    y = np.array([1, 1, 1, 0, 0, 0, 0])

    pipe.fit(X, y)
    assert pipe.predict(X).shape[0] == 7

    preprocess = FeatureUnion([("dense", lang), ("sparse", CountVectorizer())])

    assert preprocess.fit_transform(X).shape[0] == 7



import warnings
from typing import Union, List, Tuple

import numpy as np
from sklearn.metrics import pairwise_distances
from sklearn.decomposition import TruncatedSVD
from sklearn.feature_extraction.text import TfidfVectorizer

from whatlies.embedding import Embedding
from whatlies.embeddingset import EmbeddingSet
from whatlies.language._common import SklearnTransformerMixin


class TFIDFVectorLanguage(SklearnTransformerMixin):
    """
    This object is used to lazily fetch [Embedding][whatlies.embedding.Embedding]s or
    [EmbeddingSet][whatlies.embeddingset.EmbeddingSet]s from a tf/idf language
    backend. This object is meant for retreival, not plotting.

    This model will first train a scikit-learn
    [TfidfVectorizer](https://scikit-learn.org/stable/modules/generated/sklearn.feature_extraction.text.TfidfVectorizer.html)
    after which it will perform dimensionality reduction to make the numeric representation a vector. The reduction occurs via
    [TruncatedSVD](https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.TruncatedSVD.html#sklearn.decomposition.TruncatedSVD),
    also from scikit-learn.

    Warning:
        This method does not implement a word embedding in the traditional sense. The interpretation needs to
        be altered. The information that is captured here only relates to the words/characters that are used in the
        text. There is no notion of meaning that should be suggested.

        Also, in order to keep this system consistent with the rest of the api you train the system when you retreive
        vectors if you just use `__getitem__`. If you want to seperate train/test you need to call `fit_manual`
        yourself or use it in a scikit-learn pipeline.

    Arguments:
        n_components: Number of components that TruncatedSVD will reduce to.
        lowercase: If the tokens need to be lowercased beforehand.
        analyzer: Which analyzer to use, can be "word", "char", "char_wb".
        ngram_range: The range that specifies how many ngrams to use.
        min_df: Ignore terms that have a document frequency strictly lower than the given threshold.
        max_df: Ignore terms that have a document frequency strictly higher than the given threshold.
        binary: Determines if the counts are binary or if they can accumulate.
        strip_accents: Remove accents and perform normalisation. Can be set to "ascii" or "unicode".
        random_state: Random state for SVD algorithm.

    For more elaborate explainers on these arguments, check out the scikit-learn
    [documentation](https://scikit-learn.org/stable/modules/generated/sklearn.feature_extraction.text.CountVectorizer.html#sklearn.feature_extraction.text.CountVectorizer).

    **Usage**:

    ```python
    from whatlies.language import TFIDFVectorLanguage
    lang = TFIDFVectorLanguage(n_components=2, ngram_range=(1, 2), analyzer="char")
    lang[['pizza', 'pizzas', 'firehouse', 'firehydrant']]
    ```
    """

    def __init__(
        self,
        n_components: int,
        lowercase: bool = True,
        analyzer: str = "char",
        ngram_range: Tuple[int, int] = (1, 2),
        min_df: Union[int, float] = 1,
        max_df: Union[int, float] = 1.0,
        binary: bool = False,
        strip_accents: str = None,
        random_state: int = 42,
    ):
        self.n_components = n_components
        self.lowercase = lowercase
        self.analyzer = analyzer
        self.ngram_range = ngram_range
        self.min_df = min_df
        self.max_df = max_df
        self.binary = binary
        self.strip_accents = strip_accents
        self.random_state = random_state
        self.svd = TruncatedSVD(n_components=n_components, random_state=random_state)
        self.cv = TfidfVectorizer(
            lowercase=lowercase,
            analyzer=analyzer,
            ngram_range=ngram_range,
            min_df=min_df,
            max_df=max_df,
            binary=binary,
            strip_accents=strip_accents,
        )
        self.fitted_manual = False
        self.corpus = {}

    def fit_manual(self, query):
        """
        Fit the model manually. This way you can call `__getitem__` independantly of training.

        Arguments:
            query: list of strings

        **Usage**

        ```python
        from whatlies.language import CountVectorLanguage
        lang = CountVectorLanguage(n_components=2, ngram_range=(1, 2), analyzer="char")
        lang.fit_manual(['pizza', 'pizzas', 'firehouse', 'firehydrant'])
        lang[['piza', 'pizza', 'pizzaz', 'fyrehouse', 'firehouse', 'fyrehidrant']]
        ```
        """
        if any([len(q) == 0 for q in query]):
            raise ValueError(
                "You've passed an empty string to the language model which is not allowed."
            )
        X = self.cv.fit_transform(query)
        self.svd.fit(X)
        self.fitted_manual = True
        self.corpus = query
        return self

    def __getitem__(self, query: Union[str, List[str]]):
        """
        Retreive a set of embeddings.

        Arguments:
            query: list of strings

        **Usage**

        ```python
        from whatlies.language import CountVectorLanguage
        lang = CountVectorLanguage(n_components=2, ngram_range=(1, 2), analyzer="char")
        lang[['pizza', 'pizzas', 'firehouse', 'firehydrant']]
        ```
        """
        orig_str = isinstance(query, str)
        if orig_str:
            query = [query]
        if any([len(q) == 0 for q in query]):
            raise ValueError(
                "You've passed an empty string to the language model which is not allowed."
            )
        if self.fitted_manual:
            X = self.cv.transform(query)
            X_vec = self.svd.transform(X)
        else:
            X = self.cv.fit_transform(query)
            X_vec = self.svd.fit_transform(X)
        if orig_str:
            return Embedding(name=query[0], vector=X_vec[0])
        return EmbeddingSet(
            *[Embedding(name=n, vector=v) for n, v in zip(query, X_vec)]
        )

    def _prepare_queries(self, lower):
        queries = [w for w in self.corpus]
        if lower:
            queries = [w for w in queries if w.lower() == w]
        return queries

    def _calculate_distances(self, emb, queries, metric):
        vec = emb.vector
        vector_matrix = np.array([self[w].vector for w in queries])
        # there are NaNs returned, good to investigate later why that might be
        vector_matrix = np.array(
            [np.zeros(v.shape) if np.any(np.isnan(v)) else v for v in vector_matrix]
        )
        return pairwise_distances(vector_matrix, vec.reshape(1, -1), metric=metric)

    def score_similar(
        self,
        emb: Union[str, Embedding],
        n: int = 10,
        metric="cosine",
        lower=False,
    ) -> List:
        """
        Retreive a list of (Embedding, score) tuples that are the most similar to the passed query.
        Note that we will only consider words that were passed in the `.fit_manual()` step.

        Arguments:
            emb: query to use
            n: the number of items you'd like to see returned
            metric: metric to use to calculate distance, must be scipy or sklearn compatible
            lower: only fetch lower case tokens

        Returns:
            An list of ([Embedding][whatlies.embedding.Embedding], score) tuples.
        """
        if isinstance(emb, str):
            emb = self[emb]

        queries = self._prepare_queries(lower=lower)
        distances = self._calculate_distances(emb=emb, queries=queries, metric=metric)
        by_similarity = sorted(zip(queries, distances), key=lambda z: z[1])

        if len(self.corpus) < n:
            raise ValueError(
                f"You're trying to retreive {n} items while the corpus only trained on {len(self.corpus)}."
            )

        if len(queries) < n:
            warnings.warn(
                f"We could only find {len(queries)} feasible words. Consider changing `top_n` or `lower`",
                UserWarning,
            )

        return [(self[q], float(d)) for q, d in by_similarity[:n]]

    def embset_similar(
        self,
        emb: Union[str, Embedding],
        n: int = 10,
        lower=False,
        metric="cosine",
    ) -> EmbeddingSet:
        """
        Retreive an [EmbeddingSet][whatlies.embeddingset.EmbeddingSet] that are the most similar to the passed query.
        Note that we will only consider words that were passed in the `.fit_manual()` step.

        Arguments:
            emb: query to use
            n: the number of items you'd like to see returned
            metric: metric to use to calculate distance, must be scipy or sklearn compatible
            lower: only fetch lower case tokens

        Returns:
            An [EmbeddingSet][whatlies.embeddingset.EmbeddingSet] containing the similar embeddings.
        """
        embs = [
            w[0] for w in self.score_similar(emb=emb, n=n, lower=lower, metric=metric)
        ]
        return EmbeddingSet({w.name: w for w in embs})



from abc import ABC, abstractmethod

from whatlies import EmbeddingSet
from whatlies.transformers._common import new_embedding_dict


class Transformer(ABC):
    """
    This is the abstract base class for all the transformers. Each subclass of
    this class should at least implement the `fit` and `transform` methods.
    """

    def __init__(self, *args, **kwargs) -> None:
        self.is_fitted = False

    def __call__(self, embset: EmbeddingSet) -> EmbeddingSet:
        if not self.is_fitted:
            self.fit(embset)
        return self.transform(embset)

    @abstractmethod
    def fit(self, embset: EmbeddingSet) -> "Transformer":
        """
        Fit the transformer on the given `EmbeddingSet` instance (if neccessary). This method should
        set the `is_fitted` flag to `True`.

        Arguments:
            embset: an `EmbeddingSet` instance used for fitting the transformer.
        """
        raise NotImplementedError

    @abstractmethod
    def transform(self, embset: EmbeddingSet) -> EmbeddingSet:
        """
        Transform the given `EmbeddingSet` instance.

        Arguments:
            embset: an `EmbeddingSet` instance to be transformed.
        """
        raise NotImplementedError


class SklearnTransformer(Transformer):
    """
    This class is a wrapper around scikit-learn. Since many of our transformers follow
    the scikit-learn API we might be able to save a whole lot of code this way.
    """

    def __init__(self, tfm, name, *args, **kwargs):
        super().__init__()
        self.tfm = tfm(*args, **kwargs)
        self.name = name

    def fit(self, embset: EmbeddingSet) -> "SklearnTransformer":
        if not self.is_fitted:
            # This is a bit of an anti-pattern. You should not need to `self.tfm`. However, there are
            # some packages like OpenTSNE that return a different kind of estimator once an estimator has been fitted.
            self.tfm = self.tfm.fit(embset.to_X())
        self.is_fitted = True
        return self

    def transform(self, embset: EmbeddingSet) -> EmbeddingSet:
        names, X = embset.to_names_X()
        if not self.is_fitted:
            self.tfm.fit(X)
        new_vecs = self.tfm.transform(X)
        new_dict = new_embedding_dict(names, new_vecs, embset)
        return EmbeddingSet(new_dict, name=f"{embset.name}.{self.name}()")



import pytest

from whatlies.language import TFIDFVectorLanguage


@pytest.fixture
def lang():
    lang = TFIDFVectorLanguage(n_components=3, ngram_range=(1, 2), analyzer="char")
    return lang.fit_manual(
        ["pizza", "pizzas", "firehouse", "firehydrant", "cat", "dog"]
    )


def test_basic_docs_usage1(lang):
    embset = lang[["pizza", "pizzas", "firehouse", "firehydrant"]]
    assert embset.to_dataframe().shape == (4, 3)


def test_basic_docs_usage2(lang):
    embset = lang[["piza", "pizza", "pizzaz", "fyrehouse", "firehouse", "fyrehidrant"]]
    assert embset.to_dataframe().shape == (6, 3)


def test_basic_docs_usage3(lang):
    embedding = lang["piza"]
    assert embedding.name == "piza"
    assert embedding.vector.shape == (3,)


def test_score_similar(lang):
    assert len(lang.score_similar("cat", 6)) == 6


def test_value_errors1(lang):
    with pytest.raises(ValueError):
        _ = lang[""]


def test_value_errors2(lang):
    with pytest.raises(ValueError):
        lang.fit_manual(["", "cat", "dog"])


def test_retreival1(lang):
    assert lang.score_similar("doggg", n=1)[0][0].name == "dog"


def test_retreival2(lang):
    assert len(lang.score_similar("doggg", n=5)) == 5


def test_retreival_error(lang):
    with pytest.raises(ValueError):
        lang.score_similar("doggg", n=50)



import spacy
from sense2vec import Sense2Vec, Sense2VecComponent

from whatlies.embedding import Embedding
from whatlies.embeddingset import EmbeddingSet


class Sense2VecLanguage:
    """
    This object is used to lazily fetch [Embedding][whatlies.embedding.Embedding]s or
    [EmbeddingSet][whatlies.embeddingset.EmbeddingSet]s from a sense2vec language
    backend. This object is meant for retreival, not plotting.

    Arguments:
        sense2vec_path: path to downloaded vectors

    **Usage**:

    ```python
    > lang = Sense2VecLanguage(sense2vec_path="/path/to/reddit_vectors-1.1.0")
    > lang['bank|NOUN']
    > lang['bank|VERB']
    ```

    Important:
        The reddit vectors are not given by this library.
        You can find the download link [here](https://github.com/explosion/sense2vec#pretrained-vectors).

    Warning:
        This tool is temporarily not supported because sense2vec isn't supported by spaCy v3 just yet.

    """

    def __init__(self, sense2vec_path):
        self.sense2vec_path = sense2vec_path
        self.s2v = Sense2Vec().from_disk(sense2vec_path)

    def __getitem__(self, query):
        """
        Retreive a single embedding or a set of embeddings.

        Arguments:
            query: single string or list of strings

        **Usage**
        ```python
        > lang = SpacyLanguage("en_core_web_md")
        > lang['duck|NOUN']
        > lang[['duck|NOUN'], ['duck|VERB']]
        ```
        """
        if isinstance(query, str):
            vec = self.s2v[query]
            return Embedding(query, vec)
        return EmbeddingSet(*[self[tok] for tok in query])

    def embset_similar(self, query, n=10):
        """
        Retreive an [EmbeddingSet][whatlies.embeddingset.EmbeddingSet] that are the most simmilar to the passed query.

        Arguments:
            query: query to use
            n: the number of items you'd like to see returned

        Returns:
            An [EmbeddingSet][whatlies.embeddingset.EmbeddingSet] containing the similar embeddings.
        """
        return EmbeddingSet(
            *[self[tok] for tok, sim in self.s2v.most_similar(query, n=n)],
            name=f"Embset[s2v similar_{n}:{query}]",
        )

    def score_similar(self, query, n=10):
        """
        Retreive an EmbeddingSet that are the most simmilar to the passed query.

        Arguments:
            query: query to use
            n: the number of items you'd like to see returned

        Returns:
            An list of ([Embedding][whatlies.embedding.Embedding], score) tuples.
        """
        return [(self[tok], sim) for tok, sim in self.s2v.most_similar(query, n=n)]


class Sense2VecSpacyLanguage:
    """
    This object is used to lazily fetch `Embedding`s from a sense2vec language
    backend. Note that it is different than an `EmbeddingSet` in the sense
    it does not have anything precomputed.
    **Usage**:
    ```
    lang = Sense2VecLanguage(spacy_model="en_core_web_sm", sense2vec="/path/to/reddit_vectors-1.1.0")
    lang['bank|NOUN']
    lang['bank|VERB']
    ```
    """

    def __init__(self, model_name, sense2vec_path):
        self.nlp = spacy.load(model_name)
        s2v = Sense2VecComponent(self.nlp.vocab).from_disk(sense2vec_path)
        self.nlp.add(s2v)

    def __getitem__(self, string):
        doc = self.nlp(string)
        vec = doc.vector
        start, end = 0, -1
        split_string = string.split(" ")
        for idx, word in enumerate(split_string):
            if word[0] == "[":
                start = idx
            if word[-1] == "]":
                end = idx + 1
        if start != 0:
            if end != -1:
                vec = doc[start:end].vector
        return Embedding(string, vec)



import pytest
import numpy as np

from spacy.vocab import Vocab
from spacy.language import Language
from whatlies.language import SpacyLanguage


@pytest.fixture()
def color_lang():
    vector_data = {
        "red": np.array([1.0, 0.0]),
        "green": np.array([0.5, 0.5]),
        "blue": np.array([0.0, 1.0]),
        "purple": np.array([0.0, 1.0]),
    }

    vocab = Vocab(strings=list(vector_data.keys()))
    for word, vector in vector_data.items():
        vocab.set_vector(word, vector)
    nlp = Language(vocab=vocab)
    return SpacyLanguage(nlp)


def test_basic_usage(color_lang):
    queries = [
        "green is blue and yellow",
        "purple is red and blue",
        "purple isn't same as red!",
        "red and blue is a like a glue!",
    ]
    emb = color_lang[queries]
    assert len(emb) == 4
    assert emb[queries[0]].name == "green is blue and yellow"
    assert emb[queries[0]].vector.shape == (2,)


def test_score_similar_one(color_lang):
    scores = color_lang.score_similar("blue", n=2, prob_limit=None, lower=False)
    print(scores)
    assert all([s[1] == 0 for s in scores])
    assert "blue" in [s[0].name for s in scores]
    assert "purple" in [s[0].name for s in scores]


def test_single_token_words(color_lang):
    # test for issue here: https://github.com/RasaHQ/whatlies/issues/5
    assert len(color_lang["red"].vector) > 0



from ._countvector_lang import CountVectorLanguage
from ._tfidf_lang import TFIDFVectorLanguage
from ._bpemblang import BytePairLanguage
from ._gensim_lang import GensimLanguage

from whatlies.error import NotInstalled


try:
    from ._tfhub_lang import TFHubLanguage
    from ._convert_lang import ConveRTLanguage
    from ._sentence_encode_lang import UniversalSentenceLanguage
except ModuleNotFoundError:
    TFHubLanguage = NotInstalled("TFHubLanguage", "tfhub")
    ConveRTLanguage = NotInstalled("ConveRTLanguage", "tfhub")
    UniversalSentenceLanguage = NotInstalled("UniversalSentenceLanguage", "tfhub")

try:
    from ._hftransformers_lang import HFTransformersLanguage, LaBSELanguage
except ModuleNotFoundError:
    HFTransformersLanguage = NotInstalled("HFTransformersLanguage", "transformers")
    LaBSELanguage = NotInstalled("LaBSELanguage", "transformers")

try:
    from ._sense2vec_lang import Sense2VecLanguage
except ModuleNotFoundError:
    Sense2VecLanguage = NotInstalled("Sense2VecLanguage", "sense2vec")

try:
    from ._fasttext_lang import FasttextLanguage
except ModuleNotFoundError:
    FasttextLanguage = NotInstalled("FasttextLanguage", "fasttext")

try:
    from ._spacy_lang import SpacyLanguage
except ModuleNotFoundError:
    SpacyLanguage = NotInstalled("SpacyLanguage", "spacy")

try:
    from ._sentencetfm_lang import SentenceTFMLanguage
except ModuleNotFoundError:
    SentenceTFMLanguage = NotInstalled("SentenceTFMLanguage", "sentence_tfm")

try:
    from ._diet_lang import DIETLanguage
except ModuleNotFoundError:
    DIETLanguage = NotInstalled("DIETLanguage", "rasa")


__all__ = [
    "SpacyLanguage",
    "FasttextLanguage",
    "CountVectorLanguage",
    "TFIDFVectorLanguage",
    "BytePairLanguage",
    "GensimLanguage",
    "ConveRTLanguage",
    "TFHubLanguage",
    "HFTransformersLanguage",
    "UniversalSentenceLanguage",
    "SentenceTFMLanguage",
    "LaBSELanguage",
    "DIETLanguage",
]



import pytest

from whatlies.language import BytePairLanguage


@pytest.fixture()
def lang():
    return BytePairLanguage("en", vs=1000, dim=25, cache_dir="tests/cache")


def test_single_token_words(lang):
    assert lang["red"].vector.shape == (25,)
    assert len(lang[["red", "blue"]]) == 2


def test_similar_retreival(lang):
    assert len(lang.score_similar("hi", 10)) == 10
    assert len(lang.embset_similar("hi", 10)) == 10


@pytest.mark.parametrize("item", [2, 0.12341])
def test_raise_error(lang, item):
    with pytest.raises(ValueError):
        _ = lang[item]



from whatlies import Embedding, EmbeddingSet
from whatlies.language._common import SklearnTransformerMixin
from sentence_transformers import SentenceTransformer


class SentenceTFMLanguage(SklearnTransformerMixin):
    """
    This class provides the abitilty to load and use the encoding strategies found in the [sentence transformers](https://www.sbert.net/index.html) library.

    A full list of pretrained embeddings can be found [here](https://www.sbert.net/docs/pretrained_models.html).

    Arguments:
        name: name of the model to load, be sure that it's downloaded or trained beforehand

    **Usage**:

    ```python
    > from whatlies.language import SentenceTFMLanguage
    > lang = SentenceTFMLanguage('distilbert-base-nli-stsb-mean-tokens')
    > lang['python is great']
    > lang[['python is a language', 'python is a snake', 'dogs are cool animals']]
    ```
    """

    def __init__(self, name):
        self.name = name
        try:
            self.model = SentenceTransformer(name)
        except OSError:
            raise ValueError(
                f"Make sure that the name (got {name}) is found on this website:\nhttps://www.sbert.net/docs/pretrained_models.html"
            )

    def __getitem__(self, query):
        if isinstance(query, str):
            return Embedding(query, vector=self.model.encode(query))
        else:
            return EmbeddingSet(*[self[tok] for tok in query])



from typing import List, Union, Any

import numpy as np
import transformers as trf

from whatlies.embedding import Embedding
from whatlies.embeddingset import EmbeddingSet
from whatlies.language._common import SklearnTransformerMixin


class HFTransformersLanguage(SklearnTransformerMixin):
    """
    This language class can be used to load Hugging Face Transformer models and use them to obtain
    representation of input string(s) as [Embedding][whatlies.embedding.Embedding] or
    [EmbeddingSet][whatlies.embeddingset.EmbeddingSet].

    Important:
        To use this language class, either of TensorFlow or PyTorch should be installed.

        This language model does not contain a vocabulary, so it cannot be used
        to retreive similar tokens. Use an `EmbeddingSet` instead.

        This language backend might require you to manually install extra dependencies
        unless you installed via either;

        ```
        pip install whatlies[transformers]
        pip install whatlies[all]
        ```

    Arguments:
        model_name_or_path: A string which is the name or identifier of a model from
            [Hugging Face model repository](https://huggingface.co/models), or is the path to a local directory
            which contains a pre-trained transformer model files.
        kwargs: Additional key-value pair argument(s) which are passed to `transformers.pipeline` function.

    **Usage**:

    ```python
    > from whatlies.language import HFTransformersLanguage
    > lang = HFTransformersLanguage('bert-base-cased')
    > lang['today is a nice day']
    > lang = HFTransformersLanguage('gpt2')
    > lang[['day and night', 'it is as clear as day', 'today the sky is clear']]
    ```
    """

    def __init__(self, model_name_or_path: str, **kwargs: Any) -> None:
        self.model_name_or_path = model_name_or_path
        self.kwargs = kwargs
        self.model = trf.pipeline(
            task="feature-extraction", model=model_name_or_path, **kwargs
        )

    def __getitem__(self, query: Union[str, List[str]]):
        """
        Retreive a single embedding or a set of embeddings.

        Arguments:
            query: A single string or a list of strings

        Returns:
            An instance of [Embedding][whatlies.embedding.Embedding] (when `query` is a string)
            or [EmbeddingSet][whatlies.embeddingset.EmbeddingSet] (when `query` is a list of strings).
            The embedding vector is computed as the sum of hidden-state representaions of tokens
            (excluding special tokens, e.g. [CLS]).

        **Usage**

        ```python
        > from whatlies.language import HFTransformersLanguage
        > lang = HFTransformersLanguage('bert-base-cased')
        > lang['today is a nice day']
        > lang = HFTransformersLanguage('gpt2')
        > lang[['day and night', 'it is as clear as day', 'today the sky is clear']]
        ```
        """
        if isinstance(query, str):
            return self._get_embedding(query)
        return EmbeddingSet(*[self._get_embedding(q) for q in query])

    def _get_embedding(self, query: str):
        features = np.array(self.model(query, padding=False)[0])
        special_tokens_mask = self.model.tokenizer(
            query, return_special_tokens_mask=True, return_tensors="np"
        )["special_tokens_mask"][0]
        vec = features[np.logical_not(special_tokens_mask)].sum(axis=0)
        return Embedding(query, vec)


def LaBSELanguage():
    """
    Retreive a [Language Agnostic Bert](https://ai.googleblog.com/2020/08/language-agnostic-bert-sentence.html) model from huggingface.

    The model is suggested to support 109 languages. You can see the language list in the apendix of the original paper
    found [here](https://arxiv.org/pdf/2007.01852.pdf).

    Important:
        This object will automatically download a large file if it is not cached yet.

        This language model does not contain a vocabulary, so it cannot be used
        to retreive similar tokens. Use an `EmbeddingSet` instead.

        This language backend might require you to manually install extra dependencies
        unless you installed via either;

        ```
        pip install whatlies[transformers]
        pip install whatlies[all]
        ```

    **Usage**:

    ```python
    from whatlies.language import LaBSELanguage
    lang = LaBSELanguage()

    texts = ['ik vind honden leuk', 'i really like dogs', 'me gusta los perros!',
             'let us talk about money', 'laten we over geld praten', 'hablemos de dinero',
             'los stroopwafels son impresionantes', 'stroopwafels zijn heerlijk',
             'give me more stroopwafels']

    lang[texts].plot_similarity()
    ```

    ![](https://rasahq.github.io/whatlies/images/labse.png)
    """
    return HFTransformersLanguage("sentence-transformers/LaBSE")



from whatlies.transformers._pca import Pca

from whatlies.transformers._noise import Noise
from whatlies.transformers._addrandom import AddRandom
from whatlies.transformers._tsne import Tsne
from whatlies.transformers._normalizer import Normalizer
from whatlies.transformers._transformer import SklearnTransformer, Transformer
from whatlies.error import NotInstalled

try:
    from whatlies.transformers._umap import Umap
except ImportError:
    Umap = NotInstalled("Umap", "umap")


__all__ = [
    "SklearnTransformer",
    "Transformer",
    "Pca",
    "Umap",
    "Tsne",
    "Noise",
    "AddRandom",
    "Normalizer",
]



import pathlib
import warnings

from whatlies.embedding import Embedding
from whatlies.embeddingset import EmbeddingSet
from whatlies.language._common import SklearnTransformerMixin

from rasa.cli.utils import get_validated_path
from rasa.model import get_model, get_model_subdirectories
from rasa.core.interpreter import RasaNLUInterpreter
from rasa.shared.nlu.training_data.message import Message


def load_interpreter(model_dir, model):
    path_str = str(pathlib.Path(model_dir) / model)
    model = get_validated_path(path_str, "model")
    model_path = get_model(model)
    _, nlu_model = get_model_subdirectories(model_path)
    return RasaNLUInterpreter(nlu_model)


class DIETLanguage(SklearnTransformerMixin):
    """
    This language represents the transformer output of a Rasa pipeline. You can use it for debugging.

    Important:
        This language model does not contain a vocabulary, so it cannot be used
        to retreive similar tokens. Use an `EmbeddingSet` instead.

        This language backend might require you to manually install extra dependencies
        unless you installed via either;

        ```
        pip install whatlies[rasa]
        ```

        Note that we only support Rasa > 2.3.

    **Usage**:

    ```python
    from whatlies.language import DIETLanguage

    lang = DIETLanguage("path/to/model.tar.gz")
    lang[['hi', 'hello', 'greetings']]
    ```
    """

    def __init__(self, model_path):
        self.model_path = model_path
        folder, file = (
            pathlib.Path(model_path).parent,
            pathlib.Path(model_path).parts[-1],
        )
        interpreter = load_interpreter(model_dir=folder, model=file)
        self.pipeline = interpreter.interpreter.pipeline

    def __getitem__(self, item):
        """
        Retreive a single embedding or a set of embeddings. We retreive the sentence encoding that
        belongs to the entire utterance.

        Arguments:
            item: single string or list of strings

        **Usage**
        ```python
        from whatlies.language import DIETLanguage

        lang = DIETLanguage("path/to/model.tar.gz")
        lang[['hi', 'hello', 'greetings']]
        ```
        """
        if isinstance(item, str):
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=RuntimeWarning)
                msg = Message({"text": item})
                for p in self.pipeline:
                    p.process(msg)
                diagnostic_data = msg.as_dict_nlu()["diagnostic_data"]
                key_of_interest = [k for k in diagnostic_data.keys() if "DIET" in k][0]
                # It's assumed that the final token in the array here represents the __CLS__ token.
                # These are also known as the "sentence embeddings"
                tensors = diagnostic_data[key_of_interest]["text_transformed"]
                return Embedding(item, tensors[-1][-1])
        if isinstance(item, list):
            return EmbeddingSet(*[self[i] for i in item])
        raise ValueError(f"Item must be list of strings got {item}.")
